@@ -45,6 +45,8 @@ AUTH_VERSIONS_V2 = ('2.0', '2', 2)
 AUTH_VERSIONS_V3 = ('3.0', '3', 3)
 USER_METADATA_TYPE = tuple('x-%s-meta-' % type_ for type_ in
                            ('container', 'account', 'object'))
+URI_PATTERN_INFO = re.compile(r'/info')
+URI_PATTERN_VERSION = re.compile(r'\/v\d+\.?\d*(\/.*)?')
 
 try:
     from logging import NullHandler
@@ -70,6 +72,9 @@ except ImportError:
     pass
 try:
     from keystoneclient.v3 import client as ksclient_v3
+    from keystoneauth1.identity import v3
+    from keystoneauth1 import session
+    from keystoneauth1 import exceptions as ksauthexceptions
 except ImportError:
     pass
 
@@ -438,6 +443,15 @@ class HTTPConnection(object):
         if timeout:
             self.requests_args['timeout'] = timeout
 
+    def __del__(self):
+        """Cleanup resources other than memory"""
+        if self.request_session:
+            # The session we create must be closed to free up file descriptors
+            try:
+                self.request_session.close()
+            finally:
+                self.request_session = None
+
     def _request(self, *arg, **kwarg):
         """Final wrapper before requests call, to be patched in tests"""
         return self.request_session.request(*arg, **kwarg)
@@ -606,6 +620,46 @@ Auth versions 2.0 and 3 require python-keystoneclient, install it or use Auth
 version 1.0 which requires ST_AUTH, ST_USER, and ST_KEY environment
 variables to be set or overridden with -A, -U, or -K.''')
 
+    filter_kwargs = {}
+    service_type = os_options.get('service_type') or 'object-store'
+    endpoint_type = os_options.get('endpoint_type') or 'publicURL'
+    if os_options.get('region_name'):
+        filter_kwargs['attr'] = 'region'
+        filter_kwargs['filter_value'] = os_options['region_name']
+
+    if os_options.get('auth_type') == 'v3applicationcredential':
+        try:
+            v3
+        except NameError:
+            raise ClientException('Auth v3applicationcredential requires '
+                                  'python-keystoneclient>=2.0.0')
+
+        try:
+            auth = v3.ApplicationCredential(
+                auth_url=auth_url,
+                application_credential_secret=os_options.get(
+                    'application_credential_secret'),
+                application_credential_id=os_options.get(
+                    'application_credential_id'))
+            sses = session.Session(auth=auth)
+            token = sses.get_token()
+        except ksauthexceptions.Unauthorized:
+            msg = 'Unauthorized. Check application credential id and secret.'
+            raise ClientException(msg)
+        except ksauthexceptions.AuthorizationFailure as err:
+            raise ClientException('Authorization Failure. %s' % err)
+
+        try:
+            endpoint = sses.get_endpoint_data(service_type=service_type,
+                                              endpoint_type=endpoint_type,
+                                              **filter_kwargs)
+
+            return endpoint.catalog_url, token
+        except ksauthexceptions.EndpointNotFound:
+            raise ClientException(
+                'Endpoint for %s not found - '
+                'have you specified a region?' % service_type)
+
     try:
         _ksclient = ksclient.Client(
             username=user,
@@ -633,13 +687,8 @@ variables to be set or overridden with -A, -U, or -K.''')
         raise ClientException(msg)
     except ksexceptions.AuthorizationFailure as err:
         raise ClientException('Authorization Failure. %s' % err)
-    service_type = os_options.get('service_type') or 'object-store'
-    endpoint_type = os_options.get('endpoint_type') or 'publicURL'
+
     try:
-        filter_kwargs = {}
-        if os_options.get('region_name'):
-            filter_kwargs['attr'] = 'region'
-            filter_kwargs['filter_value'] = os_options['region_name']
         endpoint = _ksclient.service_catalog.url_for(
             service_type=service_type,
             endpoint_type=endpoint_type,
@@ -708,9 +757,12 @@ def get_auth(auth_url, user, key, **kwargs):
         if kwargs.get('tenant_name'):
             os_options['tenant_name'] = kwargs['tenant_name']
 
-        if not (os_options.get('tenant_name') or os_options.get('tenant_id') or
-                os_options.get('project_name') or
-                os_options.get('project_id')):
+        if os_options.get('auth_type') == 'v3applicationcredential':
+            pass
+        elif not (os_options.get('tenant_name') or
+                  os_options.get('tenant_id') or
+                  os_options.get('project_name') or
+                  os_options.get('project_id')):
             if auth_version in AUTH_VERSIONS_V2:
                 raise ClientException('No tenant specified')
             raise ClientException('No project name or project id specified.')
@@ -912,7 +964,7 @@ def post_account(url, token, headers, http_conn=None, response_dict=None,
 
 def get_container(url, token, container, marker=None, limit=None,
                   prefix=None, delimiter=None, end_marker=None,
-                  path=None, http_conn=None,
+                  version_marker=None, path=None, http_conn=None,
                   full_listing=False, service_token=None, headers=None,
                   query_string=None):
     """
@@ -926,6 +978,7 @@ def get_container(url, token, container, marker=None, limit=None,
     :param prefix: prefix query
     :param delimiter: string to delimit the queries on
     :param end_marker: marker query
+    :param version_marker: version marker query
     :param path: path query (equivalent: "delimiter=/" and "prefix=path/")
     :param http_conn: a tuple of (parsed url, HTTPConnection object),
                       (If None, it will create the conn object)
@@ -942,17 +995,20 @@ def get_container(url, token, container, marker=None, limit=None,
         http_conn = http_connection(url)
     if full_listing:
         rv = get_container(url, token, container, marker, limit, prefix,
-                           delimiter, end_marker, path, http_conn,
-                           service_token=service_token, headers=headers)
+                           delimiter, end_marker, version_marker, path=path,
+                           http_conn=http_conn, service_token=service_token,
+                           headers=headers)
         listing = rv[1]
         while listing:
             if not delimiter:
                 marker = listing[-1]['name']
             else:
                 marker = listing[-1].get('name', listing[-1].get('subdir'))
+            version_marker = listing[-1].get('version_id')
             listing = get_container(url, token, container, marker, limit,
-                                    prefix, delimiter, end_marker, path,
-                                    http_conn, service_token=service_token,
+                                    prefix, delimiter, end_marker,
+                                    version_marker, path, http_conn,
+                                    service_token=service_token,
                                     headers=headers)[1]
             if listing:
                 rv[1].extend(listing)
@@ -970,6 +1026,8 @@ def get_container(url, token, container, marker=None, limit=None,
         qs += '&delimiter=%s' % quote(delimiter)
     if end_marker:
         qs += '&end_marker=%s' % quote(end_marker)
+    if version_marker:
+        qs += '&version_marker=%s' % quote(version_marker)
     if path:
         qs += '&path=%s' % quote(path)
     if query_string:
@@ -1807,15 +1865,17 @@ class Connection(object):
         return self._retry(None, head_container, container, headers=headers)
 
     def get_container(self, container, marker=None, limit=None, prefix=None,
-                      delimiter=None, end_marker=None, path=None,
-                      full_listing=False, headers=None, query_string=None):
+                      delimiter=None, end_marker=None, version_marker=None,
+                      path=None, full_listing=False, headers=None,
+                      query_string=None):
         """Wrapper for :func:`get_container`"""
         # TODO(unknown): With full_listing=True this will restart the entire
         # listing with each retry. Need to make a better version that just
         # retries where it left off.
         return self._retry(None, get_container, container, marker=marker,
                            limit=limit, prefix=prefix, delimiter=delimiter,
-                           end_marker=end_marker, path=path,
+                           end_marker=end_marker,
+                           version_marker=version_marker, path=path,
                            full_listing=full_listing, headers=headers,
                            query_string=query_string)
 
@@ -1918,11 +1978,22 @@ class Connection(object):
                            response_dict=response_dict,
                            headers=headers)
 
-    def get_capabilities(self, url=None):
+    def _map_url(self, url):
         url = url or self.url
         if not url:
             url, _ = self.get_auth()
-        parsed = urlparse(urljoin(url, '/info'))
+        scheme, netloc, path, params, query, fragment = urlparse(url)
+        if URI_PATTERN_VERSION.search(path):
+            path = URI_PATTERN_VERSION.sub('/info', path)
+        elif not URI_PATTERN_INFO.search(path):
+            if path.endswith('/'):
+                path += 'info'
+            else:
+                path += '/info'
+        return urlunparse((scheme, netloc, path, params, query, fragment))
+
+    def get_capabilities(self, url=None):
+        parsed = urlparse(self._map_url(url))
         if not self.http_conn:
             self.http_conn = self.http_connection(url)
         return get_capabilities((parsed, self.http_conn[1]))
